searchNodes=[{"doc":"","ref":"Alchemy.html","title":"Alchemy","type":"module"},{"doc":"","ref":"Alchemy.Experiment.html","title":"Alchemy.Experiment","type":"module"},{"doc":"Adds a candidate function to the experiment. The candidate needs to be wrapped in a function in order to be lazily-evaluated. When the experiment is run the candidate will be evaluated and compared to the control.","ref":"Alchemy.Experiment.html#candidate/2","title":"Alchemy.Experiment.candidate/2","type":"function"},{"doc":"Adds a clean function. This function will be run for each observation and can be used to reduce noise when publishing results.","ref":"Alchemy.Experiment.html#clean/2","title":"Alchemy.Experiment.clean/2","type":"function"},{"doc":"Adds a comparator to use when comparing the candidate to the control. By default the comparator is: fn ( control , candidate ) -&gt; control == candidate end","ref":"Alchemy.Experiment.html#comparator/2","title":"Alchemy.Experiment.comparator/2","type":"function"},{"doc":"Adds a control function to the experiment. Controls should be wrapped in a function in order to be lazily-evaluated","ref":"Alchemy.Experiment.html#control/2","title":"Alchemy.Experiment.control/2","type":"function"},{"doc":"","ref":"Alchemy.Experiment.html#control_value/1","title":"Alchemy.Experiment.control_value/1","type":"function"},{"doc":"Generates a new experiment. Alias for Experiment.new/1","ref":"Alchemy.Experiment.html#experiment/1","title":"Alchemy.Experiment.experiment/1","type":"function"},{"doc":"Adds an ignore clause to the experiment. This clause will be called if there is a mismatch between the control values and the candidate values. This is a useful function if you want to ignore known mismatches. Multiple ignore clauses can be stacked together.","ref":"Alchemy.Experiment.html#ignore/2","title":"Alchemy.Experiment.ignore/2","type":"function"},{"doc":"Sets the function to use for publishing results. Can accept either a module or a function. If a module name is passed in then the module is expected to have a publish/1 function which will be used for publishing results.","ref":"Alchemy.Experiment.html#publisher/2","title":"Alchemy.Experiment.publisher/2","type":"function"},{"doc":"Runs the experiment. If the candidate is provided then it will be run against the control . The control must be provided for the experiment to be run. The control is always returned. The execution order is randomized to account for any ordering issues. Execution for these functions is done in serial. This is because adding concurrent execution adds additional layers of error handling that would need to occur and doing so is best done in the users application code and not in this library code.","ref":"Alchemy.Experiment.html#run/1","title":"Alchemy.Experiment.run/1","type":"function"},{"doc":"","ref":"Alchemy.Observation.html","title":"Alchemy.Observation","type":"module"},{"doc":"","ref":"Alchemy.Observation.html#run/2","title":"Alchemy.Observation.run/2","type":"function"},{"doc":"","ref":"Alchemy.Result.html","title":"Alchemy.Result","type":"module"},{"doc":"","ref":"Alchemy.Result.html#ignored?/1","title":"Alchemy.Result.ignored?/1","type":"function"},{"doc":"","ref":"Alchemy.Result.html#matched?/1","title":"Alchemy.Result.matched?/1","type":"function"},{"doc":"","ref":"Alchemy.Result.html#mismatched?/1","title":"Alchemy.Result.mismatched?/1","type":"function"},{"doc":"","ref":"Alchemy.Result.html#new/3","title":"Alchemy.Result.new/3","type":"function"},{"doc":"","ref":"Alchemy.Result.html#raised?/1","title":"Alchemy.Result.raised?/1","type":"function"},{"doc":"Safely perform refactoring experiments in production.","ref":"readme.html","title":"Alchemy","type":"extras"},{"doc":"def deps do [ { :alchemy , &quot;~&gt; 0.2.0&quot; } ] end","ref":"readme.html#installation","title":"Alchemy - Installation","type":"extras"},{"doc":"Lets say that you have a controller that returns a list of users, and you want to change how that list of users is fetched from the database. Unit tests can help you but they can only account for examples that you've currently considered. Alchemy allows you to test your new code live in production. defmodule MyApp.UserController do alias Alchemy.Experiment def index ( conn ) do users = Experiment . new ( &quot;users-query&quot; ) |&gt; Experiment . control ( &amp; old_slow_query / 0 ) |&gt; Experiment . candidate ( &amp; new_fast_query / 0 ) |&gt; Experiment . run render ( conn , &quot;index.json&quot; , users : users ) end defp old_query do # ... end defp new_query do # ... end end Both the control and the candidate are randomized. Once all of the observations have been made the control is returned so that its easy to continue pipelining with other functions. Along with this Alchemy does a few other things: Compares the results of the control against the candidates looking for mismatches Measures the execution time of both the control and the candidates Captures and records any errors thrown by the candidates Publishes all of these results Publish results The experiment is now running but the results aren't being published. In order to publish the results We can pass a publisher function to our experiment. defmodule MyApp.UserController do require Logger alias Alchemy . { Experiment , Result } def index ( conn ) do users = Experiment . new ( &quot;users-query&quot; ) |&gt; Experiment . control ( &amp; old_slow_query / 0 ) |&gt; Experiment . candidate ( &amp; new_fast_query / 0 ) |&gt; Experiment . publisher ( &amp; publish / 1 ) |&gt; Experiment . run render ( conn , &quot;index.json&quot; , users : users ) end def publish ( result = % Alchemy.Result { } ) do name = result . name control = result . control candidate = hd ( result . candidates ) matched = Result . matched? ( result ) Logger . debug &quot;&quot;&quot; Test: \#{ name } Match?: \#{ ! Result . mismatched? ( result ) } Control - value: \#{ control . value } | duration: \#{ control . duration } Candidate - value: \#{ candidate . value } | duration: \#{ candidate . duration } &quot;&quot;&quot; end end If you want to share your publishing logic across multiple experiments then we can pass a module as the producer. Alchemy will assume that there is a publish/1 function available on the specified module. def index ( conn ) do users = Experiment . new ( &quot;users-query&quot; ) |&gt; Experiment . control ( &amp; old_slow_query / 0 ) |&gt; Experiment . candidate ( &amp; new_fast_query / 0 ) |&gt; Experiment . publisher ( Publisher ) |&gt; Experiment . run end defmodule Publisher do alias Alchemy.Result def publish ( %{ name : name } = result ) do Statix . timing ( &quot;alchemy. \#{ name } .control.duration.ms&quot; , result . control . duration ) candidate_duration = result . candidates |&gt; Enum . at ( 0 ) |&gt; Map . get ( :duration ) Statix . timing ( &quot;alchemy. \#{ name } .candidate.duration.ms&quot; , candidate_duration ) # and counts for match/ignore/mismatch: cond do Result . matched? ( result ) -&gt; Statix . increment ( &quot;alchemy. \#{ name } .matched.total&quot; ) Result . ignored? ( result ) -&gt; Statix . increment ( &quot;alchemy. \#{ name } .ignored.total&quot; ) true -&gt; Statix . increment ( &quot;alchemy. \#{ name } .mismatched.total&quot; ) store_result ( result ) end end # Store final results in a public ets table for analysis defp store_result ( result ) do payload = %{ name : result . name , control : observation_payload ( result . control ) , candidate : observation_payload ( Enum . at ( result . candidates , 0 ) ) } :ets . insert ( :alchemy , { result . name , payload } ) end # If this observation raised an error then store the error # otherwise store the &quot;cleaned&quot; value. defp observation_payload ( observation ) do cond do Result . raised? ( observation ) -&gt; %{ error : observation . error , stacktrace : observation . stacktrace } true -&gt; %{ value : observation . cleaned_value } end end end Multiple Candidates It's possible to create experiments with multiple candidates: def some_query do experiment ( &quot;test multiple candidates&quot; ) |&gt; control ( &amp; old_query / 0 ) |&gt; candidate ( &amp; foo_query / 0 ) |&gt; candidate ( &amp; bar_query / 0 ) |&gt; candidate ( &amp; baz_query / 0 ) |&gt; run end Comparing results By default alchemy compares results with == . You can override this by supplying your own comparator: def user_name do experiment ( &quot;test name is correct&quot; ) |&gt; control ( fn -&gt; %{ id : 1 , name : &quot;Alice&quot; } end ) |&gt; candidate ( fn -&gt; %{ id : 2 , name : &quot;Alice&quot; } end ) |&gt; comparator ( fn ( control , candidate ) -&gt; control . name == candidate . name end ) |&gt; run end Cleaning values Often you won't want to publish the full value that is returned from the observation. For instance if you're comparing 2 large structs you may just want to store the id fields so you can analysize them later. You can pass a cleaner function to your experiment to define how these cleaned values are created (by default the cleaner is an identity function). def users do experiment ( &quot;user-list&quot; ) |&gt; control ( fn -&gt; Repo . all ( User ) end ) |&gt; candidate ( fn -&gt; UserContext . list ( ) end ) |&gt; clean ( fn user -&gt; user . id end ) |&gt; run end Ignoring results If you have certain scenarios that you know will always result in mismatches then you can ignore them so that they don't end up in your published results. Multiple ignore clauses can be stacked together. def staff? ( id ) do user = User . get ( id ) experiment ( &quot;staff?&quot; ) |&gt; control ( fn -&gt; old_role_check ( user ) end ) |&gt; candidate ( fn -&gt; Roles . staff? ( user ) end ) |&gt; ignore ( fn control , candidate -&gt; # If the control passed but the candidate didn&#39;t check to see if its because # the user has no email. We haven&#39;t implemented that in our new system yet. control &amp;&amp; ! candidate &amp;&amp; user . email == &quot;&quot; end ) # Admin users are always considered staff. |&gt; ignore ( fn _ , _ , -&gt; Roles . admin? ( user ) end ) end","ref":"readme.html#perform-some-experiments","title":"Alchemy - Perform some experiments","type":"extras"},{"doc":"Alchemy tries to be as transparent as possible. Because of this all errors raised in either the control or the candidate are rescued and stored in the result. This allows you to see any errors in your candidates or control when you publish your results. Once the results have been published the control value is returned. If the control raised an error during its execution that error will be reraised with its original stacktrace.","ref":"readme.html#exception-handling","title":"Alchemy - Exception handling","type":"extras"},{"doc":"Executing the control, candidates, and publishing are all done sequentially within the calling process. Trying to run these operations concurrently introduces multiple failure conditions and possibility for timeouts. These failure scenarios tend to surprise users and breaks one of Alchemy's core goals of being transparent and making refactors safe. If you want to run any of these operations concurrently then its best to do that in your own code so that you aren't surprised and can handle failures in the way that makes the most sense for your application.","ref":"readme.html#execution-details","title":"Alchemy - Execution details","type":"extras"},{"doc":"Practically all of the concepts and many of the examples are shamelessly stolen from Github's amazing Scientist project. All credit goes to them for popularizing this idea and creating such a smart api.","ref":"readme.html#prior-art","title":"Alchemy - Prior Art","type":"extras"},{"doc":"Pull Requests and Issues are greatly appreciated!","ref":"readme.html#contributing","title":"Alchemy - Contributing","type":"extras"}]